#ifdef HBMODE

#ifndef HBWRAP_H
#define HBWRAP_H
#define LOGICAL_STRICT

#include<vector>
#include<iostream>
#include<cstdlib>


///////////////////////////////////////////////////
///// Higgs Bounds Raw routines  //////////////////
///////////////////////////////////////////////////

/// Initialization wrapped routines
extern "C" {void initialize_higgsbounds_(int & nHzero, int & nHplus, char * whichanalyses);}

extern "C" {void initialize_higgsbounds_int_(int & nHzero, int & nHplus, int & whichanalyses);}

///Input routines
//Neutral sector
extern "C" {void higgsbounds_neutral_input_effc_(double * Mh, double * MhGammaTot, double * g2hjss_s, double * g2hjss_p, double * g2hjcc_s, double * g2hjcc_p, double * g2hjbb_s, double * g2hjbb_p, double * g2hjtoptop_s, double * g2hjtoptop_p,double * g2hjmumu_s, double * g2hjmumu_p, double * g2hjtautau_s,double * g2hjtautau_p, double * g2hjWW, double * g2hjZZ,double * g2hjZga, double * g2hjgaga, double * g2hjgg,double * g2hjggZ, double * g2hjhiZ, double * BR_hjinvisible,double * BR_hjhihi);}


extern "C" {void higgsbounds_neutral_input_part_(double* Mh, double* MhGammaTot, int* CP, double* CS_lep_hjZ_ratio, double* CS_lep_bbhj_ratio, double* CS_lep_tautauhj_ratio, double* CS_lep_hjhi_ratio,double* CS_gg_hj_ratio, double* CS_bb_hj_ratio,double* CS_bg_hjb_ratio, double* CS_ud_hjWp_ratio,double* CS_cs_hjWp_ratio, double* CS_ud_hjWm_ratio,double* CS_cs_hjWm_ratio, double* CS_gg_hjZ_ratio,double* CS_dd_hjZ_ratio, double* CS_uu_hjZ_ratio,double* CS_ss_hjZ_ratio, double* CS_cc_hjZ_ratio,double* CS_bb_hjZ_ratio, double* CS_tev_vbf_ratio,double* CS_tev_tthj_ratio, double* CS_lhc7_vbf_ratio,double* CS_lhc7_tthj_ratio, double* CS_lhc8_vbf_ratio,double* CS_lhc8_tthj_ratio, double* BR_hjss, double* BR_hjcc,double* BR_hjbb, double* BR_hjmumu, double* BR_hjtautau,double* BR_hjWW, double* BR_hjZZ, double* BR_hjZga,double* BR_hjgaga, double* BR_hjgg, double* BR_hjinvisible,double * BR_hjhihi);}

extern "C" {void higgsbounds_neutral_input_hadr_(double* Mh, double* GammaTot,int* CP, double* CS_lep_hjZ_ratio, double* CS_lep_bbhj_ratio,double* CS_lep_tautauhj_ratio, double* CS_lep_hjhi_ratio,double* CS_tev_hj_ratio, double* CS_tev_hjb_ratio,double* CS_tev_hjW_ratio, double* CS_tev_hjZ_ratio,double* CS_tev_vbf_ratio, double* CS_tev_tthj_ratio,double* CS_lhc7_hj_ratio, double* CS_lhc7_hjb_ratio,double* CS_lhc7_hjW_ratio, double* CS_lhc7_hjZ_ratio,double* CS_lhc7_vbf_ratio, double* CS_lhc7_tthj_ratio,double* CS_lhc8_hj_ratio, double* CS_lhc8_hjb_ratio,double* CS_lhc8_hjW_ratio, double* CS_lhc8_hjZ_ratio,double* CS_lhc8_vbf_ratio, double* CS_lhc8_tthj_ratio,double* BR_hjss, double* BR_hjcc, double* BR_hjbb,double* BR_hjmumu, double* BR_hjtautau, double* BR_hjWW,double* BR_hjZZ, double* BR_hjZga, double* BR_hjgaga,double* BR_hjgg, double* BR_hjinvisible, double* BR_hjhihi);}

//Charged sector
extern "C" {void higgsbounds_charged_input_(double * MHplus, double * MHplusGammaTot,double* CS_lep_HpjHmi_ratio, double* BR_tWpb,double* BR_tHpjb, double* BR_Hpjcs, double* BR_Hpjcb,double* BR_Hptaunu);}

//SLHA optional input with a file
extern "C" {void higgsbounds_input_slha_(char * SLHAfilename);}

//optional use of mass uncertainties both for neutral and charged higgses (vectors of uncertainties to be passed)
extern "C" {void higgsbounds_set_mass_uncertainties(double * dMhneut, double * dMhch);}

//Routine to actually run Higgs Bounds
extern "C" {void run_higgsbounds_(int & HBresult, int & chan, double & obsratio,int & ncombined);}

///Finishing wrapped routines
extern "C" {void finish_higgsbounds_();}


/////////////////////////////////////////////////
//// Wrapper class to make interface simpler ////
/////////////////////////////////////////////////

class HBobject{
 public:
  HBobject(){};
  HBobject(int nHzero_in,int nHplus_in,int whichanalyses_in,int whichinput_in):nHzero(nHzero_in),nHplus(nHplus_in),whichanalyses(whichanalyses_in),whichinput(whichinput_in){};
  void rewrite_values(int nHzero_in,int nHplus_in,int whichanalyses_in,int whichinput_in);
  void InitialiseHiggsBounds(); //Constructor (allocates memory and sets variables)
  ~HBobject(void); //Destructor
  void run(int & HBresult, int & chan, double & obsratio,int & ncombined); //Run HiggsBounds
  
  //////////////////////////////////////////////////////////////////////////////
  ///////////// All possible vectors and matrices that can be passed to HB /////

  //Hadronic TypeInput variables
  double * Mh; //All objects of this type are vectors
  double * MhGammaTot;
  int * CP; 
  double * CS_lep_hjZ_ratio;
  double * CS_lep_bbhj_ratio;
  double * CS_lep_tautauhj_ratio;
  double ** CS_lep_hjhi_ratio; // THIS IS A MATRIX
  double * CS_tev_hj_ratio;
  double * CS_tev_hjb_ratio;
  double * CS_tev_hjW_ratio;
  double * CS_tev_hjZ_ratio;
  double * CS_tev_vbf_ratio; 
  double * CS_tev_tthj_ratio;
  double * CS_lhc7_hj_ratio;
  double * CS_lhc7_hjb_ratio;
  double * CS_lhc7_hjW_ratio;
  double * CS_lhc7_hjZ_ratio;
  double * CS_lhc7_vbf_ratio;
  double * CS_lhc7_tthj_ratio;
  double * CS_lhc8_hj_ratio;
  double * CS_lhc8_hjb_ratio;
  double * CS_lhc8_hjW_ratio;
  double * CS_lhc8_hjZ_ratio;
  double * CS_lhc8_vbf_ratio;
  double * CS_lhc8_tthj_ratio;
  double * BR_hjss;
  double * BR_hjcc;
  double * BR_hjtt;
  double * BR_hjbb;
  double * BR_hjmumu;
  double * BR_hjtautau;
  double * BR_hjWW;
  double * BR_hjZZ;
  double * BR_hjZga;
  double * BR_hjgaga;
  double * BR_hjgg;
  double * BR_hjinvisible;
  double ** BR_hjhihi; //THIS IS A MATRIX!!!
  
  //Partonic type input variables, which are not also above
  double * CS_gg_hj_ratio;
  double * CS_bb_hj_ratio;
  double * CS_bg_hjb_ratio;
  double * CS_ud_hjWp_ratio;
  double * CS_cs_hjWp_ratio;
  double * CS_ud_hjWm_ratio;
  double * CS_cs_hjWm_ratio;
  double * CS_gg_hjZ_ratio;
  double * CS_dd_hjZ_ratio;
  double * CS_uu_hjZ_ratio;
  double * CS_ss_hjZ_ratio;
  double * CS_cc_hjZ_ratio;
  double * CS_bb_hjZ_ratio;
  
  //effective coupling variables
  double * g2hjss_s;
  double * g2hjss_p;
  double * g2hjcc_s;
  double * g2hjcc_p;
  double * g2hjbb_s;
  double * g2hjbb_p;
  double * g2hjtoptop_s;
  double * g2hjtoptop_p;
  double * g2hjmumu_s;
  double * g2hjmumu_p;
  double * g2hjtautau_s;
  double * g2hjtautau_p;
  double * g2hjWW;
  double * g2hjZZ;
  double * g2hjZga;
  double * g2hjgaga;
  double * g2hjgg;
  double * g2hjggZ;
  double ** g2hjhiZ;//THIS IS A MATRIX!!!
  
  //Charged Higgses
  double * MHplus;
  double * MHplusGammaTot;
  double * CS_lep_HpjHmi_ratio;
  double * BR_tWpb;
  double * BR_tHpjb;
  double * BR_Hpjcs;
  double * BR_Hpjcb;
  double * BR_Hptaunu;
  
  int nHzero,nHplus,whichanalyses,whichinput; //Initialisation
 private:
  void allocate_arrays(void); //Dynamical memory allocation of arrays
  void set_f90_matrices(void); //Set f90 matrices to zero
  void input(void); //Pass all input to HB. This is done automatically when run is called


  double * CS_lep_hjhi_ratio_f90;
  double * g2hjhiZ_f90; //This is the matrix that is actually passed!!!
  double * BR_hjhihi_f90; //This is the matrix that is actually passed!!!
};


#endif

#endif
