#ifndef SCANNERSANALYSISUTILS_H
#define SCANNERSANALYSISUTILS_H
#include <iostream>
#include <fstream>
#include <vector>
#include <cmath>
#include <cstdlib>
#include <gsl/gsl_fit.h>
#include <gsl/gsl_multifit.h>
#include <gsl/gsl_statistics_double.h>

#ifdef HBMODE
#include <map>
#include "ExtInterfaces/HBWrap.h"
#include "ScannerS_exp_values.h"
#endif

bool fileExists(const char *fileName);

//Class to generate user defined tables to speed up runs
class TableData{
 public:
  double operator()(std::vector<double> & Point); //Call Table to get interpolated value
  void create_from_scratch(std::vector<double> & MinRanges_in,std::vector<double> & MaxRanges_in,std::vector<int> & NsGrid_in,double (*func_in)(std::vector<double> & Point, void *), void *parfunc_inc); // build a new table from scratch using a functions and given ranges
  void create_from_scratch_flush(std::vector<double> & MinRanges_in,std::vector<double> & MaxRanges_in,std::vector<int> & NsGrid_in,double (*func_in)(std::vector<double> & Point, void *), void *parfunc_inc,const std::string fileName); // build a new table from scratch using a functions and given ranges
  void clear_data(void); //resets all member objects which hold the data, to be empty
  void calcpartition(int A,int N,std::vector<int> & indices);
  void calcpartition_unit(int A,int N,std::vector<int> & indices);
  int calc_grid_coord(std::vector<int> & indices);
  void calcNprodCoeffs();
  void save_table(const char *fileName);
  void create_from_file(const char *fileName);
  friend std::istream & operator>>(std::istream & is, TableData & T); // Interface to input table from a stream
  friend std::ostream & operator<<(std::ostream & os, const TableData & T); // Interface to output table to a stream
  double Min(int i);
  double Max(int i);
  int Ngrid(int i);

 private:
  double (*func)(std::vector<double> & Point,void *); //Function to build table (if bulding from scratch)
  std::vector<double> grid; //Array to store table data
  std::vector<double> MinRanges,MaxRanges; //Ranges for the table
  std::vector<double> unit_cell_sizes; 
  std::vector<int> NsGrid; //number of points along each direction on the grid.
  std::vector<int> NprodCoeff;
  void * parfunc;
};

// Least square fits
double fit_multi_params(double * Xij_in, double * y_in, int npoints, double *coeffs, int ncoeffs, double * err);

//signal strength check
#ifdef HBMODE
bool CheckSignalStrengths_gen(const int nsigma, const double sep, const HBobject & HBA,std::map<std::string,double> & results);
#endif

#endif
