#include "ScannerSN2HDM.h"

using namespace std;

string N2HDM::filepath="ScannerScore/ScannerSModels/N2HDM/N2HDM.potential";
const double N2HDM::Ecm2 = 2000.; //todo: make these global
const double N2HDM::Ecm7 = 7000.;
const double N2HDM::Ecm8 = 8000.;
int N2HDM::fail_global=0;
int N2HDM::fail_STU=0;
int N2HDM::fail_Z_bbar=0;
int N2HDM::fail_bs_gamma=0;
int N2HDM::fail_HB=0;
int N2HDM::fail_rates=0;


N2HDM::N2HDM(){
//CONSTRUCTOR

#ifdef HBMODE
  HiggsBoundsAnalysis.rewrite_values(nHzero,nHplus,whichanalyses,whichinput); //sets the values needed to initialize HiggsBounds (will be initialized in UserInitCalcs)
#endif

  //Loading numerical tables for cross-sections
  Load_ggF_bbF_noEW_SusHiPP_PPbarTables_LHC_plus_Tevatron();
  //Load numerical Superiso tables
  TableSuperisoBsgammaType1.create_from_file("ScannerScore/data/SuperisoBsgamma2HDMType1.dat");
  TableSuperisoBsgammaType2.create_from_file("ScannerScore/data/SuperisoBsgamma2HDMType2.dat");
  TableSuperisoBsgammaType3.create_from_file("ScannerScore/data/SuperisoBsgamma2HDMType3.dat");
  TableSuperisoBsgammaType4.create_from_file("ScannerScore/data/SuperisoBsgamma2HDMType4.dat");

  //Initialise arrays to zero

  //Higgs couplings to vector bosons
  Rhvv.fill(0);
  for(auto& r : Rhhv)
    r.fill(0);
}

void N2HDM::readfile(std::istream & in){
//This function is called from the overloaded operator >> when a N2HDM input file is provided.
//On the base class, Potential::readfile simply calls openinput() on "in".
//since N2HDM files only contain the relevant variables, "this" readfile opens an auxiliary file in the same format as Potential::readfile
//and then replaces the relevante variables bty the values in "in".
  ifstream modelfile;
  OpenModelFile(modelfile,filepath);
  openinput(modelfile);

  mass_bounds[0]=getvariable(in,Cp.Mass.Vecmin[0],Cp.Mass.Vecmax[0],">m1");
  mass_bounds[1]=getvariable(in,Cp.Mass.Vecmin[1],Cp.Mass.Vecmax[1],">m2");
  mass_bounds[2]=getvariable(in,Cp.Mass.Vecmin[2],Cp.Mass.Vecmax[2],">m3");
  mass_bounds[3]=getvariable(in,Cp.Mass.Vecmin[3],Cp.Mass.Vecmax[3],">mA");
  mass_bounds[4]=getvariable(in,Cp.Mass.Vecmin[4],Cp.Mass.Vecmax[4],">mH+");
  mass_bounds[5]=getvariable(in,Cp.Mass.Vecmin[5],Cp.Mass.Vecmax[5],">mH+");
  L_bounds[11]=getvariable(in,Cp.L.Vecmin[11],Cp.L.Vecmax[11],">m12^2");
  // bool alpha_bounds=getvariable(in,MixingPars.Vecmin[0],MixingPars.Vecmax[0],">alpha");
  // bool delta_bounds=getvariable(in,MixingPars.Vecmin[1],MixingPars.Vecmax[1],">delta");
  // getvariable(in,MixingPars.Vecmin[2],MixingPars.Vecmax[2],">gamma");
  bool tanbeta_bounds=getvariable(in,PhiPar.Vecmin[0],PhiPar.Vecmax[0],">tanbeta");
  bool vs_bounds=getvariable(in,PhiPar.Vecmin[1],PhiPar.Vecmax[1],">vs");

  getvariable(in,Type2HDM,">2HDM_type:");

  //check of all independent parameters were defined in the input file
  for(int i=0;i!=6;++i)
    if(!mass_bounds[i]){
      cerr<<"ScannerS ERROR: Unable to find a necessary parameter in the 2HDM input file. This model ALWAYS requires you define the range for mH, mh, mS, mA, mH+, alpha, delta, tanbeta, vs and m12^2. These are independent parameters that will be randomly generated."<<endl;
      exit(EXIT_FAILURE);
    }
  if((!L_bounds[11]) || (!tanbeta_bounds) ||(!vs_bounds)){
    cerr<<"ScannerS ERROR: Unable to find a necessary parameter in the 2HDM input file. This model ALWAYS requires you define the range for mH, mh, mA, mH+, alpha, delta, tanbeta, vs and m12^2. These are independent parameters that will be randomly generated."<<endl;
    exit(EXIT_FAILURE);
  }

  //check if there were bounds imposed on the dependent couplings
  L_bounds[0]=getvariable(in,Cp.L.Vecmin[0],Cp.L.Vecmax[0],">m11^2");
  L_bounds[1]=getvariable(in,Cp.L.Vecmin[1],Cp.L.Vecmax[1],">m22^2");
  L_bounds[2]=getvariable(in,Cp.L.Vecmin[2],Cp.L.Vecmax[2],">lambda5");
  L_bounds[3]=getvariable(in,Cp.L.Vecmin[3],Cp.L.Vecmax[3],">lambda1");
  L_bounds[4]=getvariable(in,Cp.L.Vecmin[4],Cp.L.Vecmax[4],">lambda2");
  L_bounds[5]=getvariable(in,Cp.L.Vecmin[5],Cp.L.Vecmax[5],">lambda3");
  L_bounds[6]=getvariable(in,Cp.L.Vecmin[6],Cp.L.Vecmax[6],">lambda4");
  L_bounds[7]=getvariable(in,Cp.L.Vecmin[7],Cp.L.Vecmax[7],">ms^2");
  L_bounds[8]=getvariable(in,Cp.L.Vecmin[8],Cp.L.Vecmax[8],">lambda6");
  L_bounds[9]=getvariable(in,Cp.L.Vecmin[9],Cp.L.Vecmax[9],">lambda7");
  L_bounds[10]=getvariable(in,Cp.L.Vecmin[10],Cp.L.Vecmax[10],">lambda8");
  //check if the singlet admixture for the input states is bounded
  sing_bounds[0]=getvariable(in,sing_min[0],sing_max[0],">sing1");
  sing_bounds[1]=getvariable(in,sing_min[1],sing_max[1],">sing2");
  sing_bounds[2]=getvariable(in,sing_min[2],sing_max[2],">sing3");
  //check if the minimum mass seperation to the charged Higgs is bounded
  double temp;
  deltamHc_bound = getvariable(in,temp,deltamHc_max,">delta(mHc)");
}

std::string N2HDM::name() const{
//returns the name of the type of potential used.
  string name("N2HDM");
  return name;
}

void N2HDM::PrintHeader() const{

  std::clog<<endl<<"***********************************************"<<endl;
  std::clog<<"*********** "<<scanners_version<<" ************"<<endl;
  std::clog<<"***********************************************" <<endl;
  std::clog<<endl<<"---------- SUMMARY INFO FOR THIS RUN ---------- "<<endl;
  std::clog<<endl<<"Potential Type: N2HDM"<<endl;
  std::clog<<endl<<"V(Phi1,Phi2) = L0Phi1'Phi1 + L1Phi2'Phi2 + L11(Phi1'Phi2+Phi2'Phi1) + (L3/2)(Phi1'Phi1)² + (L4/2)(Phi2'Phi2)² + L5Phi1'Phi1Phi2'Phi2 + L6Phi2'Phi1Phi1'Phi2 + (L2/2)((Phi1'Phi2)²+(Phi2'Phi1)²) + (L7/2) PhiS^2 + (L8/8)PhiS^4 + (L9/2)PhiS^2(Phi1'Phi1) + (L10/2)PhiS^2(Phi2'Phi2)"<<endl;
  std::clog<<endl<< "The expanded potential is:"<<endl;
  std::clog<<endl<<(*this)<<endl;
  std::clog<<endl<<"The ranges of parameters to be scanned are:"<<endl;
  std::clog<<endl<<"The 2HDM Type is "<<Type2HDM<<endl;
  std::clog<<endl<<"Ranges for vacuum expectation values:"<<endl;
  for(size_t i=0;i!=Phi.size();++i){
    if(phi_bounds.at(i))
      std::clog<<"phi"<<i<<" in ["<<Phi.vgetmin(i)<<":"<<Phi.vgetmax(i)<<"]"<< endl;
  }
  std::clog<<"phi8 in ["<<PhiPar.Vecmin[1]<<":"<<PhiPar.Vecmax[1]<<"]"<< endl;
  std::clog<<endl<<"Ranges for couplings:"<< endl;
  for(size_t i=0;i!=Cp.L.size();++i){
    if(L_bounds.at(i))
      std::clog<<"L"<<i<<" in ["<<Cp.L.vgetmin(i)<<":"<<Cp.L.vgetmax(i)<<"]"<< endl;
  }
  std::clog<<endl<<"Ranges for physical state masses:"<< endl;
  for(size_t i=0;i!=mass_bounds.size();++i){
    if(mass_bounds[i])
    std::clog<<"m"<<i<<" in ["<<Cp.Mass.vgetmin(i)<<":"<<Cp.Mass.vgetmax(i)<<"]"<<endl;
  }
  std::clog<<endl<<"Range for tanbeta (v2/v6 parameterization):"<<endl;
  std::clog<<"tanbeta in ["<<PhiPar.Vecmin[0]<<":"<<PhiPar.Vecmax[0]<<"]"<<endl;

  std::clog<<endl<<"Allowed singlet admixture to the input mass states:"<< endl;
  for(size_t i=0;i!=sing_bounds.size();++i){
    if(sing_bounds[i])
    std::clog<<"singlet admixture of state "<<i<<" in ["<<sing_min[i]<<":"<<sing_max[i]<<"]"<<endl;
  }

  if(deltamHc_bound)
   std::clog<<endl<<"Maximum allowed seperation of "<<deltamHc_max<<" GeV between mHc and the nearest neutral Higgs";

  std::clog<<endl<<"The seed you are using for the random number generator is: "<<r.getseed()<< endl<< endl;

}

bool N2HDM::CheckBounds() const{
//unlike Potential::CheckBounds(), this only checks dependent parameters for which the range was explicity declared in the 2HDM input file. if they were not declared in input file (and were loaded of 2HDM.potential by default) L_bounds will be false and the parameters will not be checked
  for(size_t i=0;i!=Cp.L.size();++i)//L_bounds size checked at PrintHeader
    if(L_bounds[i] && (Cp.L[i]<Cp.L.vgetmin(i) || Cp.L[i]>Cp.L.vgetmax(i))){
#ifdef VERBOSE
    std::clog<<endl<<"---> ScannerS: Point rejected by N2HDM::CheckBounds()."<<endl;
#endif
      return false;
  }
  for(size_t i=0;i!=sing_bounds.size();++i)
    if(sing_bounds[i] && ( (Mmixing[i][8]*Mmixing[i][8])*100 < sing_min[i] || (Mmixing[i][8]*Mmixing[i][8])*100 > sing_max[i])){
#ifdef VERBOSE
    std::clog<<endl<<"---> ScannerS: Point rejected by N2HDM::CheckBounds()."<<endl;
#endif
      return false;
  }


  // sanity check of the found mixing
  if(Cp.Mass[3]<Cp.Mass.Vecmin[3] ||
      fabs(Cp.Mass[4]-Cp.Mass[5])>1e-3 ||
      Cp.Mass[0] < Cp.Mass.Vecmin[0] ||
      Cp.Mass[1] < Cp.Mass.Vecmin[1] ||
      Cp.Mass[2] < Cp.Mass.Vecmin[2]) {
    clog<<"Warning in N2HDM::CheckBounds illegal or singular mixing input recieved!"<<endl;
    return false;
  }

  if(deltamHc_bound && ( std::min({abs(Cp.Mass[0]-Cp.Mass[4]),abs(Cp.Mass[1]-Cp.Mass[4]),abs(Cp.Mass[2]-Cp.Mass[4]),abs(Cp.Mass[3]-Cp.Mass[4])}) > deltamHc_max))
    return false;

  return true;
}

#ifdef HBMODE
void N2HDM::PrepareHiggsBoundsInput(){

  //Call of Hdecay to compute all branching ratios
  //todo: move to classes??
  CreateInputFileHdecayN2HDM(Type2HDM, tanbeta, a1, a2, a3, mH1, mH2, mH3, mA, mHc, Cp.L[11], Phi[8]);
  HdecayCalcN2HDM(HdecayH1,HdecayH2,HdecayH3,HdecayA,HdecayHc,HdecayTop);
  //calculate all needed angle trigonometrics
  double ca1 = cos(a1);
  double sa1 = sin(a1);
  double ca2 = cos(a2);
  double sa2 = sin(a2);
  double ca3 = cos(a3);
  double sa3 = sin(a3);

  double cb = cos(beta);
  double sb = sin(beta);

  double ca1mb = cos(a1-beta);
  double sa1mb = sin(a1-beta);


  //Defining Higgs couplings to fermions.
  array<double,3> cfd1,cfd2;
  cfd1[0] = (ca1*ca2)/cb;
  cfd1[1] = (-(ca3*sa1) - ca1*sa2*sa3)/cb;
  cfd1[2] = (-(ca1*ca3*sa2) + sa1*sa3)/cb;
  cfd2[0] = (ca2*sa1)/sb;
  cfd2[1] = (ca1*ca3 - sa1*sa2*sa3)/sb;
  cfd2[2] = (-(ca3*sa1*sa2) - ca1*sa3)/sb;

  // Rhf[iHiggs][ifermion]: iHiggs=0,1,2,3 (A,H1,H2,H3), ifermion= 0,1,2 (up-type quarks, d-type quarks and leptons)
  if(Type2HDM==1){	//Type I
    for(int i=1;i!=nHzero;i++){
  	  Rhf[i][0] = cfd2[i-1];
  	  Rhf[i][1] = cfd2[i-1];
  	  Rhf[i][2] = cfd2[i-1];
    }
    Rhf[0][0] = 1./tanbeta;
  	Rhf[0][1] = -1./tanbeta;
  	Rhf[0][2] = -1./tanbeta;
  }
  if(Type2HDM==2){	//Type II
    for(int i=1;i!=nHzero;i++){
  	  Rhf[i][0] = cfd2[i-1];
  	  Rhf[i][1] = cfd1[i-1];
  	  Rhf[i][2] = cfd1[i-1];
    }
    Rhf[0][0] = 1./tanbeta;
    Rhf[0][1] = tanbeta;
    Rhf[0][2] = tanbeta;
  }
  if(Type2HDM==3){	//Lepton Specific
    for(int i=1;i!=nHzero;i++){
  	  Rhf[i][0] = cfd2[i-1];
  	  Rhf[i][1] = cfd2[i-1];
  	  Rhf[i][2] = cfd1[i-1];
    }
    Rhf[0][0] = 1./tanbeta;
    Rhf[0][1] = -1/tanbeta;
    Rhf[0][2] = tanbeta;
  }
  if(Type2HDM==4){	//Flipped
    for(int i=1;i!=nHzero;i++){
  	  Rhf[i][0] = cfd2[i-1];
  	  Rhf[i][1] = cfd1[i-1];
  	  Rhf[i][2] = cfd2[i-1];
    }
    Rhf[0][0] = 1./tanbeta;
    Rhf[0][1] = tanbeta;
    Rhf[0][2] = -1/tanbeta;
 }

  //hiVV couplings, initialised to 0
  Rhvv[1] = ca1mb*ca2;
  Rhvv[2] = -(ca3*sa1mb) - ca1mb*sa2*sa3;
  Rhvv[3] = -(ca1mb*ca3*sa2) + sa1mb*sa3;


  //hihjV couplings, initialised to 0
  Rhhv[0][1] = -(ca2*sa1mb);
  Rhhv[0][2] = -(ca1mb*ca3) + sa1mb*sa2*sa3;
  Rhhv[0][3] = ca3*sa1mb*sa2 + ca1mb*sa3;
  for(int i=1;i!=nHzero;++i)Rhhv[i][0]=Rhhv[0][i];



  /////////////////////////////////////////////////////////
  //Passing couplings, masses and decay properties to HB //
  /////////////////////////////////////////////////////////
  //Light Higgs - H1
  HiggsBoundsAnalysis.Mh[1]=mH1;
  HiggsBoundsAnalysis.CP[1]=1;
  //Branching ratios and total width
  HiggsBoundsAnalysis.BR_hjss[1] = HdecayH1.at("BR(H1 -> s sbar)");
  HiggsBoundsAnalysis.BR_hjcc[1] = HdecayH1.at("BR(H1 -> c cbar)");
  HiggsBoundsAnalysis.BR_hjbb[1] = HdecayH1.at("BR(H1 -> b bbar)");
  HiggsBoundsAnalysis.BR_hjmumu[1] = HdecayH1.at("BR(H1 -> mu+ mu-)");
  HiggsBoundsAnalysis.BR_hjtautau[1] = HdecayH1.at("BR(H1 -> tau+ tau-)");
  HiggsBoundsAnalysis.BR_hjWW[1] = HdecayH1.at("BR(H1 -> W+ W-)");
  HiggsBoundsAnalysis.BR_hjZZ[1] = HdecayH1.at("BR(H1 -> Z Z)");
  HiggsBoundsAnalysis.BR_hjZga[1]= HdecayH1.at("BR(H1 -> Z gamma)");
  HiggsBoundsAnalysis.BR_hjgaga[1] = HdecayH1.at("BR(H1 -> gamma gamma)");
  HiggsBoundsAnalysis.BR_hjgg[1] = HdecayH1.at("BR(H1 -> g g)");
  HiggsBoundsAnalysis.BR_hjhihi[1][0] = HdecayH1.at("BR(H1 -> A A)");
  HiggsBoundsAnalysis.MhGammaTot[1] = HdecayH1.at("Width");

  //Medium Higgs - H2
  HiggsBoundsAnalysis.Mh[2]=mH2;
  HiggsBoundsAnalysis.CP[2]=1;
  //Branching ratios and total width
  HiggsBoundsAnalysis.BR_hjss[2] = HdecayH2.at("BR(H2 -> s sbar)");
  HiggsBoundsAnalysis.BR_hjcc[2] = HdecayH2.at("BR(H2 -> c cbar)");
  HiggsBoundsAnalysis.BR_hjbb[2] = HdecayH2.at("BR(H2 -> b bbar)");
  HiggsBoundsAnalysis.BR_hjmumu[2] = HdecayH2.at("BR(H2 -> mu+ mu-)");
  HiggsBoundsAnalysis.BR_hjtautau[2] = HdecayH2.at("BR(H2 -> tau+ tau-)");
  HiggsBoundsAnalysis.BR_hjWW[2] = HdecayH2.at("BR(H2 -> W+ W-)");
  HiggsBoundsAnalysis.BR_hjZZ[2] = HdecayH2.at("BR(H2 -> Z Z)");
  HiggsBoundsAnalysis.BR_hjZga[2]= HdecayH2.at("BR(H2 -> Z gamma)");
  HiggsBoundsAnalysis.BR_hjgaga[2] = HdecayH2.at("BR(H2 -> gamma gamma)");
  HiggsBoundsAnalysis.BR_hjgg[2] = HdecayH2.at("BR(H2 -> g g)");
  HiggsBoundsAnalysis.BR_hjhihi[2][0] = HdecayH2.at("BR(H2 -> A A)");
  HiggsBoundsAnalysis.BR_hjhihi[2][1] = HdecayH2.at("BR(H2 -> H1 H1)");
  HiggsBoundsAnalysis.MhGammaTot[2] = HdecayH2.at("Width");

  //Heavy Higgs - H3
  HiggsBoundsAnalysis.Mh[3]=mH3;
  HiggsBoundsAnalysis.CP[3]=1;
  //Branching ratios and total width
  HiggsBoundsAnalysis.BR_hjss[3] = HdecayH3.at("BR(H3 -> s sbar)");
  HiggsBoundsAnalysis.BR_hjcc[3] = HdecayH3.at("BR(H3 -> c cbar)");
  HiggsBoundsAnalysis.BR_hjbb[3] = HdecayH3.at("BR(H3 -> b bbar)");
  HiggsBoundsAnalysis.BR_hjmumu[3] = HdecayH3.at("BR(H3 -> mu+ mu-)");
  HiggsBoundsAnalysis.BR_hjtautau[3] = HdecayH3.at("BR(H3 -> tau+ tau-)");
  HiggsBoundsAnalysis.BR_hjWW[3] = HdecayH3.at("BR(H3 -> W+ W-)");
  HiggsBoundsAnalysis.BR_hjZZ[3] = HdecayH3.at("BR(H3 -> Z Z)");
  HiggsBoundsAnalysis.BR_hjZga[3]= HdecayH3.at("BR(H3 -> Z gamma)");
  HiggsBoundsAnalysis.BR_hjgaga[3] = HdecayH3.at("BR(H3 -> gamma gamma)");
  HiggsBoundsAnalysis.BR_hjgg[3] = HdecayH3.at("BR(H3 -> g g)");
  HiggsBoundsAnalysis.BR_hjhihi[3][0] = HdecayH3.at("BR(H3 -> A A)");
  HiggsBoundsAnalysis.BR_hjhihi[3][1] = HdecayH3.at("BR(H3 -> H1 H1)");
  HiggsBoundsAnalysis.BR_hjhihi[3][2] = HdecayH3.at("BR(H3 -> H2 H2)");
  HiggsBoundsAnalysis.MhGammaTot[3] = HdecayH3.at("Width");

  //Pseudo scalar - A
  HiggsBoundsAnalysis.Mh[0]=mA;
  HiggsBoundsAnalysis.CP[0]=-1;
  //Branching ratios and total width
  HiggsBoundsAnalysis.BR_hjss[0] = HdecayA.at("BR(A -> s sbar)");
  HiggsBoundsAnalysis.BR_hjcc[0] = HdecayA.at("BR(A -> c cbar)");
  HiggsBoundsAnalysis.BR_hjbb[0] = HdecayA.at("BR(A -> b bbar)");
  HiggsBoundsAnalysis.BR_hjmumu[0] = HdecayA.at("BR(A -> mu+ mu-)");
  HiggsBoundsAnalysis.BR_hjtautau[0] = HdecayA.at("BR(A -> tau+ tau-)");
  HiggsBoundsAnalysis.BR_hjWW[0] = 0.;
  HiggsBoundsAnalysis.BR_hjZZ[0] = 0.;
  HiggsBoundsAnalysis.BR_hjZga[0]= HdecayA.at("BR(A -> Z gamma)");
  HiggsBoundsAnalysis.BR_hjgaga[0] = HdecayA.at("BR(A -> gamma gamma)");
  HiggsBoundsAnalysis.BR_hjgg[0] = HdecayA.at("BR(A -> g g)");
  HiggsBoundsAnalysis.MhGammaTot[0] = HdecayA.at("Width");

  //Charged Higgs - H^+
  HiggsBoundsAnalysis.MHplus[0] = mHc;
  HiggsBoundsAnalysis.BR_tWpb[0] = HdecayTop.at("BR(t -> b W+)");
  HiggsBoundsAnalysis.BR_tHpjb[0] = HdecayTop.at("BR(t -> b H+)");
  HiggsBoundsAnalysis.BR_Hpjcs[0] = HdecayHc.at("BR(H+ -> c sbar)");
  HiggsBoundsAnalysis.BR_Hpjcb[0] = HdecayHc.at("BR(H+ -> c bbar)");
  HiggsBoundsAnalysis.BR_Hptaunu[0] = HdecayHc.at("BR(H+ -> tau+ nu_tau)");
  HiggsBoundsAnalysis.MHplusGammaTot[0] = HdecayHc.at("Width");

  //Neutral Higgs hadronic input
  for(size_t i=0;i!=nHzero;++i){
    //LEP
    HiggsBoundsAnalysis.CS_lep_hjZ_ratio[i]=pow(Rhvv[i],2);
    HiggsBoundsAnalysis.CS_lep_bbhj_ratio[i]=pow(Rhf[i][1],2);
    HiggsBoundsAnalysis.CS_lep_tautauhj_ratio[i]=pow(Rhf[i][2],2);
    for (size_t j=0; j!=nHzero;++j)
      HiggsBoundsAnalysis.CS_lep_hjhi_ratio[i][j]=pow(Rhhv[i][j],2);
    //Tevatron
    HiggsBoundsAnalysis.CS_tev_hj_ratio[i]=(InternalTableSusHiPPbar_ggF_noEW(Ecm2,HiggsBoundsAnalysis.Mh[i],HiggsBoundsAnalysis.CP[i],Rhf[i][0],Rhf[i][1])+InternalTableSusHiPPbar_bbF_noEW(Ecm2,HiggsBoundsAnalysis.Mh[i],HiggsBoundsAnalysis.CP[i],Rhf[i][1]))/InternalTableSusHiPPbar_ggF_noEW(Ecm2,HiggsBoundsAnalysis.Mh[i],1,1,1);
    HiggsBoundsAnalysis.CS_tev_hjb_ratio[i]=pow(Rhf[i][1],2);
    HiggsBoundsAnalysis.CS_tev_hjW_ratio[i]=pow(Rhvv[i],2);
    HiggsBoundsAnalysis.CS_tev_hjZ_ratio[i]=pow(Rhvv[i],2);
    HiggsBoundsAnalysis.CS_tev_vbf_ratio[i]=pow(Rhvv[i],2);
    HiggsBoundsAnalysis.CS_tev_tthj_ratio[i]=pow(Rhf[i][0],2);
    //LHC7
    HiggsBoundsAnalysis.CS_lhc7_hj_ratio[i]=(InternalTableSusHiPP_ggF_noEW(Ecm7,HiggsBoundsAnalysis.Mh[i],HiggsBoundsAnalysis.CP[i],Rhf[i][0],Rhf[i][1])+InternalTableSusHiPP_bbF_noEW(Ecm7,HiggsBoundsAnalysis.Mh[i],HiggsBoundsAnalysis.CP[i],Rhf[i][1]))/InternalTableSusHiPP_ggF_noEW(Ecm7,HiggsBoundsAnalysis.Mh[i],1,1,1);
    HiggsBoundsAnalysis.CS_lhc7_hjb_ratio[i]=pow(Rhf[i][1],2);
    HiggsBoundsAnalysis.CS_lhc7_hjW_ratio[i]=pow(Rhvv[i],2);
    HiggsBoundsAnalysis.CS_lhc7_hjZ_ratio[i]=pow(Rhvv[i],2);
    HiggsBoundsAnalysis.CS_lhc7_vbf_ratio[i]=pow(Rhvv[i],2);
    HiggsBoundsAnalysis.CS_lhc7_tthj_ratio[i]=pow(Rhf[i][0],2);
    //LHC8
    HiggsBoundsAnalysis.CS_lhc8_hj_ratio[i]=(InternalTableSusHiPP_ggF_noEW(Ecm8,HiggsBoundsAnalysis.Mh[i],HiggsBoundsAnalysis.CP[i],Rhf[i][0],Rhf[i][1])+InternalTableSusHiPP_bbF_noEW(Ecm8,HiggsBoundsAnalysis.Mh[i],HiggsBoundsAnalysis.CP[i],Rhf[i][1]))/InternalTableSusHiPP_ggF_noEW(Ecm8,HiggsBoundsAnalysis.Mh[i],1,1,1);
    HiggsBoundsAnalysis.CS_lhc8_hjb_ratio[i]=pow(Rhf[i][1],2);
    HiggsBoundsAnalysis.CS_lhc8_hjW_ratio[i]=pow(Rhvv[i],2);
    HiggsBoundsAnalysis.CS_lhc8_hjZ_ratio[i]=pow(Rhvv[i],2);
    HiggsBoundsAnalysis.CS_lhc8_vbf_ratio[i]=pow(Rhvv[i],2);
    HiggsBoundsAnalysis.CS_lhc8_tthj_ratio[i]=pow(Rhf[i][0],2);
  }

	//Charged Higgs hadronic input
  HiggsBoundsAnalysis.CS_lep_HpjHmi_ratio[0] = 1;
}

bool N2HDM::CheckSignalStrengths() {
  //check at two sigma
  constexpr int nsigma = 2;
  //with a seperation of 5GeV
  constexpr double seperation = 5.;

  return CheckSignalStrengths_gen(nsigma,seperation,HiggsBoundsAnalysis,mu_results);
}

#endif

bool N2HDM::TestZbbarbLEP() const{
  // if (Type2HDM==1 || Type2HDM==3)return true;
  return tanbeta>Interp(mHc,CKMfitterZbbar);
}

bool N2HDM::Test_bs_gamma(){
  //flavour Physics observables
  const double bsgamma_max=4.1;
  const double bsgamma_min=3.;
  vector<double>massbeta = {mHc,beta};
  if(Type2HDM==1)
    bsgamma=TableSuperisoBsgammaType1(massbeta);  ////Call superiso table to test B-> X_s gamma
  else if(Type2HDM==2){
    if(mHc < 480)return false; //Misiak et al. 2015
    bsgamma=TableSuperisoBsgammaType2(massbeta);  ////Call superiso table to test B-> X_s gamma
  }
  else if(Type2HDM==3)
    bsgamma=TableSuperisoBsgammaType3(massbeta);  ////Call superiso table to test B-> X_s gamma
  else if(Type2HDM==4){
    if(mHc < 480)return false; //Misiak et al. 2015
    bsgamma=TableSuperisoBsgammaType4(massbeta);  ////Call superiso table to test B-> X_s gamma
  }
  if(bsgamma*1e4>bsgamma_max || bsgamma*1e4 <bsgamma_min)
      return false;
  return true;
}

void N2HDM::PhysicalStatesN2HDM(){
  //This routines returns the masses of the scalars in the 2HDM as well as the alpha mixing angle for the 2 CP even Higgs bosons and the beta angle between the vevs.
  //	INPUT:
  //       Mass - internal (vector like) array of masses
  //       Mixing - internal (matrix like) mixing matrix
  //       Phi - internal (vector like) array of field values
  //	OUTPUT:
  //       mHc - Mass of the charged Higgs
  //       mA - Mass of the CP odd Higgs
  //    - mass ordering
  //       mH1 - lightest CP even Higgs
  //       mH2 - medium mass CP even Higgs
  //       mH3 - heaviest CP even Higgs
  //       a1,a2,a3 mixing angles in mass ordering
  //    - singlet ordering
  //       singlet admixtures sH1,sH2,sH3

  //Computing beta from VEVs
  tanbeta=Phi[6]/Phi[2];
  beta=atan(tanbeta);

  mA=Cp.Mass[3]; //Mass of the pseudo scalar (numbering checked in VERBOSE mode)
  mHc=Cp.Mass[4]; //Mass of the charged Higgs (numbering checked in VERBOSE mode. Note Mass[4] is the same because it corresponds to the imaginary part).

  unsigned int h[3]={0,1,2};
  int temp;

  if(Cp.Mass[h[0]]>Cp.Mass[h[2]]){
    temp=h[2];
    h[2]=h[0];
    h[0]=temp;
  }
  if(Cp.Mass[h[0]]>Cp.Mass[h[1]]){
    temp=h[1];
    h[1]=h[0];
    h[0]=temp;
  }
  else if(Cp.Mass[h[1]]>Cp.Mass[h[2]]){
    temp=h[2];
    h[2]=h[1];
    h[1]=temp;
  }

  mH1 = Cp.Mass[h[0]];
  mH2 = Cp.Mass[h[1]];
  mH3 = Cp.Mass[h[2]];

  //clog<<"Order: "<<h[0]<<"; "<<h[1]<<"; "<<h[2]<<endl;

  //todo: Eigen?
  double det=-Mmixing[h[0]][8]*Mmixing[h[1]][6]*Mmixing[h[2]][2] + Mmixing[h[0]][6]*Mmixing[h[1]][8]*Mmixing[h[2]][2] + Mmixing[h[0]][8]*Mmixing[h[1]][2]*Mmixing[h[2]][6] - Mmixing[h[0]][2]*Mmixing[h[1]][8]* Mmixing[h[2]][6] - Mmixing[h[0]][6]*Mmixing[h[1]][2]*Mmixing[h[2]][8] + Mmixing[h[0]][2]*Mmixing[h[1]][6]*Mmixing[h[2]][8];

  // clog<<"original Mixing Matrix reordered\n";
  // clog<<"{{"<<Mmixing[h[0]][2]<<","<<Mmixing[h[0]][6]<<","<<Mmixing[h[0]][8]<<"},{";
  // clog<<Mmixing[h[1]][2]<<","<<Mmixing[h[1]][6]<<","<<Mmixing[h[1]][8]<<"},{";
  // clog<<Mmixing[h[2]][2]<<","<<Mmixing[h[2]][6]<<","<<Mmixing[h[2]][8]<<"}}"<<endl;
  // clog<<"det= "<<det<<endl;

  // all angles shall be constrained to [-Pi/2,+Pi/2],
  // which means that the elements containing only cos need to be positive
  // to ensure the proper determinant,
  // either an even or odd number of line sign changes (=reflect_) is needed
  std::array<int,3>reflect {1,1,1};
  if(det>0){
    //clog<<"positive determinant"<<endl;
    if(Mmixing[h[0]][2]<0 && Mmixing[h[2]][8]<0){
      reflect[0]=-1;
      reflect[2]=-1;
    }
    else if(Mmixing[h[0]][2]<0){
      reflect[0]=-1;
      reflect[1]=-1;
    }
    else if(Mmixing[h[2]][8]<0){
      reflect[1]=-1;
      reflect[2]=-1;
    }
  }
  else{
    if(Mmixing[h[0]][2]<0 && Mmixing[h[2]][8]<0){
      reflect[0]=-1;
      reflect[1]=-1;
      reflect[2]=-1;
    }
    else if(Mmixing[h[0]][2]<0){
      reflect[0]=-1;
    }
    else if(Mmixing[h[2]][8]<0){
      reflect[2]=-1;
    }
    else{//both positive
      reflect[1]=-1;
    }
  }

  //compute the alphas from the mixing matrix
  a2=asin(reflect[0]*Mmixing[h[0]][8]);
  a1=asin(reflect[0]*Mmixing[h[0]][6]/cos(a2));
  a3=asin(reflect[1]*Mmixing[h[1]][8]/cos(a2));
  //compute the singlet admixtures from the mixing matrix
  sH1 = pow(Mmixing[h[0]][8],2);
  sH2 = pow(Mmixing[h[1]][8],2);
  sH3 = pow(Mmixing[h[2]][8],2);

}


bool N2HDM::CheckTreeUni(){
  constexpr double pi = acos(-1);
  maxev = std::max({
    abs(1/2.*(Cp.L[3]+Cp.L[4]-sqrt(4*Cp.L[2]*Cp.L[2] + pow(Cp.L[3]-Cp.L[4],2)))),
    abs(1/2.*(Cp.L[3]+Cp.L[4]+sqrt(4*Cp.L[2]*Cp.L[2] + pow(Cp.L[3]-Cp.L[4],2)))),
    abs(Cp.L[5]-Cp.L[6]),
    abs(Cp.L[5]+2*Cp.L[6]-3*Cp.L[2]),
    abs(Cp.L[5]+2*Cp.L[6]+3*Cp.L[2]),
    abs(Cp.L[9]),
    abs(Cp.L[10]),
    abs(1/2.*(Cp.L[3]+Cp.L[4]-sqrt(pow(Cp.L[4]-Cp.L[4],2)+4*Cp.L[6]*Cp.L[6]))),
    abs(1/2.*(Cp.L[3]+Cp.L[4]+sqrt(pow(Cp.L[4]-Cp.L[4],2)+4*Cp.L[6]*Cp.L[6])))
  });
  maxev /= pi;
  if(maxev > 8)
    return false;

  //numeric case
  double * coeffs;
  coeffs = new double [3+1];
  double * x;
  x = new double [2*3];

  coeffs[0]=(
    -108*Cp.L[3]*Cp.L[4]*Cp.L[8]
    +48*Cp.L[5]*Cp.L[5]*Cp.L[8]
    +48*Cp.L[5]*Cp.L[6]*Cp.L[8]
    +12*Cp.L[6]*Cp.L[6]*Cp.L[8]
    +24*Cp.L[4]*Cp.L[9]*Cp.L[9]
    -32*Cp.L[5]*Cp.L[9]*Cp.L[10]
    -16*Cp.L[6]*Cp.L[9]*Cp.L[10]
    +24*Cp.L[3]*Cp.L[10]*Cp.L[10]
  );
  coeffs[1]=(
    36*Cp.L[3]*Cp.L[4]
    -16*Cp.L[5]*Cp.L[5]
    -16*Cp.L[5]*Cp.L[6]
    -4*Cp.L[6]*Cp.L[6]
    +18*Cp.L[3]*Cp.L[8]
    +18*Cp.L[4]*Cp.L[8]
    -4*Cp.L[9]*Cp.L[9]
    -4*Cp.L[10]*Cp.L[10]
  );
  coeffs[2]=-6*Cp.L[3]-6*Cp.L[4]-3*Cp.L[8];
  coeffs[3]=1.;
  polyroots(coeffs,x,3);
  for(size_t i=0;i!=3;i++){
    if(abs(x[2*i+1])>abs(1e-8*x[2*i])){
      clog<<"ScannerS Warning in N2HDM::CheckTreeUni(): Complex S matrix EV "<<x[2*i]/2./pi<<" + i * "<<x[2*i+1]/2./pi<<" encountered, dropping the point.\n";
      return false;
    }
    if(x[2*i]/2./pi > maxev)
      maxev = x[2*i]/2./pi;
  }
  if(maxev > 8)
    return false;
}

bool N2HDM::TestSTU() {
  if(CheckVirtual(mA,mHc) &&
      CheckVirtual(mA,mH1) &&
      CheckVirtual(mHc,mHc) &&
      CheckVirtual(mHc,mH1)) {

    //
    double cb = cos(beta);
    double sb = sin(beta);
    double ca1 = cos(a1);
    double ca2 = cos(a2);
    double ca3 = cos(a3);
    double sa1 = sin(a1);
    double sa2 = sin(a2);
    double sa3 = sin(a3);

    // feed values into the STU input
    // matrices
    STU_N2HDM.m_U.at(0).at(0) = cb;
    STU_N2HDM.m_U.at(0).at(1) = -sb;
    STU_N2HDM.m_U.at(1).at(0) = sb;
    STU_N2HDM.m_U.at(1).at(1) = cb;
    STU_N2HDM.m_Vr.at(0).at(0) = 0;
    STU_N2HDM.m_Vr.at(0).at(1) = 0;
    STU_N2HDM.m_Vr.at(0).at(2) = ca1*ca2;
    STU_N2HDM.m_Vr.at(0).at(3) = -(ca3*sa1) - ca1*sa2*sa3;
    STU_N2HDM.m_Vr.at(0).at(4) = -(ca1*ca3*sa2) + sa1*sa3;
    STU_N2HDM.m_Vr.at(1).at(0) = 0;
    STU_N2HDM.m_Vr.at(1).at(1) = 0;
    STU_N2HDM.m_Vr.at(1).at(2) = ca2*sa1;
    STU_N2HDM.m_Vr.at(1).at(3) = ca1*ca3 - sa1*sa2*sa3;
    STU_N2HDM.m_Vr.at(1).at(4) = -(ca3*sa1*sa2) - ca1*sa3;
    STU_N2HDM.m_Vi.at(0).at(0) = cb;
    STU_N2HDM.m_Vi.at(0).at(1) = -sb;
    STU_N2HDM.m_Vi.at(0).at(2) = 0;
    STU_N2HDM.m_Vi.at(0).at(3) = 0;
    STU_N2HDM.m_Vi.at(0).at(4) = 0;
    STU_N2HDM.m_Vi.at(1).at(0) = sb;
    STU_N2HDM.m_Vi.at(1).at(1) = cb;
    STU_N2HDM.m_Vi.at(1).at(2) = 0;
    STU_N2HDM.m_Vi.at(1).at(3) = 0;
    STU_N2HDM.m_Vi.at(1).at(4) = 0;
    //masses
    STU_N2HDM.mHzero.at(0)=mA;
    STU_N2HDM.mHzero.at(1)=mH1;
    STU_N2HDM.mHzero.at(2)=mH2;
    STU_N2HDM.mHzero.at(3)=mH3;
    STU_N2HDM.mHplus.at(0)=mHc;

    //evaluate
    bool result = STU_N2HDM.evaluate();
    // clog<<"S "<<STU_N2HDM.S;
    // clog<<" T "<<STU_N2HDM.T;
    // clog<<" U "<<STU_N2HDM.U<<endl;
    return result;
  } else {
    return false;
    // alternative behaviour for test runs
    // std::clog<<"Warning in N2HDM::TestSTU: Attempt to aplly STU test for too low masses. Test ignored for this point."<<endl;
    // STU_N2HDM.S = 0;
    // STU_N2HDM.T = 0;
    // STU_N2HDM.U = 0;
    // return true;
  }
}
