#include "RxSM.h"

using namespace std;

string RxSM::filepath="ScannerScore/ScannerSModels/RxSM/RxSM.potential";
int RxSM::nHzero;
int RxSM::nHplus=0;
int RxSM::whichanalyses=3;	//which Higgs bounds analysis
int RxSM::whichinput=1; 	//which type of input you will provide (1=hadronic)

// CONSTRUCTOR
RxSM::RxSM(void){
  
  //heavy particle masses
  for(int i=0;i!=2;++i)
    mass_bounds[i]=false;
  //goldstone masses
  for(int i=2;i!=5;++i)
    mass_bounds[i]=true;
  //couplings
  for(int i=0;i!=5;++i)
    L_bounds[i]=false;
  //VeV's: 0 to 3 are Higgs' (0,1 and 3 being fixed at 0) and 4:5 are the new Singlet's
  for(size_t i=0;i!=5;++i)
    phi_bounds[i]=true;
}

std::string RxSM::name() const{
//returns the name of the type of potential used.  
  string name("RxSM");	
  return name;
}

void RxSM::readfile(std::istream & in){
//This function is called from the overloaded operator >> when a TwoRxSM input file is provided.
//On the base class, Potential::readfile simply calls openinput() on "in".
//Since TwoHDM files only contain the relevante variables, "this" readfile opens an auxiliary file in the same format has Potential::readfile
//and then replaces the relevante variables bty the values in "in".
  ifstream modelfile;
  OpenModelFile(modelfile,filepath);
  openinput(modelfile);
  
  //VeVs
  phi_bounds[2]=getvariable(in,Phi.Vecmin[2],Phi.Vecmax[2],">vH");
  phi_bounds[4]=getvariable(in,Phi.Vecmin[4],Phi.Vecmax[4],">vS");
  //masses (actually - MLpars)
  mass_bounds[0]=getvariable(in,MLpars.Vecmin[0],MLpars.Vecmax[0],">massH");
  mass_bounds[1]=getvariable(in,MLpars.Vecmin[1],MLpars.Vecmax[1],">massS");
  //coupling
  L_bounds[0]=getvariable(in,Cp.L.Vecmin[0],Cp.L.Vecmax[0],">m^2");
  L_bounds[1]=getvariable(in,Cp.L.Vecmin[1],Cp.L.Vecmax[1],">lambda");
  L_bounds[2]=getvariable(in,Cp.L.Vecmin[2],Cp.L.Vecmax[2],">lambdaHS");
  L_bounds[3]=getvariable(in,Cp.L.Vecmin[3],Cp.L.Vecmax[3],">mS^2");
  L_bounds[4]=getvariable(in,Cp.L.Vecmin[4],Cp.L.Vecmax[4],">lambdaS");
  
  //always independent
  for(int i=0;i!=5;++i){
    if(!phi_bounds[i]){
      cerr<<"ScannerS ERROR: Unable to find a necessary parameter in the RxSM input file. This model ALWAYS requires you define the range for vH, vS, massH and massS. Aditionally you need to provide ranges for mS^2 and lambdaS in the DM case. These parameters are independent parameters that will be randomly generated."<<endl;
      for(size_t j=0;j!=5;++j)
	cerr<<"phi["<<j<<"]="<<phi_bounds[j]<<endl;
      exit(EXIT_FAILURE);
    }
  }
  for(int i=0;i!=5;++i){
    if(!mass_bounds[i]){
      cerr<<"ScannerS ERROR: Unable to find a necessary parameter in the RxSM input file. This model ALWAYS requires you define the range for vH, vS, massH, massS and mA, lambda_SH. Aditionally you need to provide ranges for mS^2 and lambdaS in the DM case). These parameters are independent parameters that will be randomly generated."<<endl;
      for(size_t j=0;j!=5;++j)
	cerr<<"mass["<<j<<"]="<<mass_bounds[j]<<endl;
      exit(EXIT_FAILURE);
    }
  }
  //DM phase
  if(Phi.Vecmin[4]==0 && Phi.Vecmax[4]==0){
    DM_candidate=true;
    nHzero=1;
    if((!L_bounds[3]) || (!L_bounds[4])){
      cerr<<"ScannerS ERROR: Unable to find a necessary parameter in the RxSM input file. This model ALWAYS requires you define the range for vH, vS, massH, massS and mA, lambda_SH. Aditionally you need to provide ranges for mS^2 and lambdaS in the DM case). These parameters are independent parameters that will be randomly generated."<<endl;
      for(size_t j=0;j!=5;++j)
	cerr<<"L["<<j<<"]="<<L_bounds[j]<<endl;
      exit(EXIT_FAILURE);
    }
  }
  else{	//broken phase
    DM_candidate=false;
    nHzero=2;
  }
#ifdef HBMODE
  HiggsBoundsAnalysis.rewrite_values(nHzero,nHplus,whichanalyses,whichinput); //sets the values needed to initialize HiggsBounds (will be initialized in UserInitCalcs)
#endif
}

void RxSM::compute_alpha(void){
  if(DM_candidate)
    alpha=0;
  else{
    int reflect1=1,reflect2=1;
    double det=Mmixing[h[0]][2]*Mmixing[h[1]][4] - Mmixing[h[0]][4]*Mmixing[h[1]][2];
    if(det>0){
      if(Mmixing[h[0]][2]<0){
	reflect1=-1;
	reflect2=-1;
      }
    }
    else{
      if(Mmixing[h[0]][2]>0)
	reflect2=-1;
      else
	reflect1=-1;
    }
    //clog<<"det: "<<det<<endl;
    
    //compute the alphas from the mixing matrix
    alpha=asin(reflect1*Mmixing[h[0]][4]);
    
    /*double diff=abs(reflect1*Mmixing[h[0]][2]-cos(alpha))+abs(reflect1*Mmixing[h[0]][4]-sin(alpha))+abs(reflect2*Mmixing[h[1]][2]+sin(alpha))+abs(reflect2*Mmixing[h[1]][4]-cos(alpha));
    if(diff>1e-14){
      clog<<"Mixing Matrix after correcting for a possible reflection is:"<<endl;
      clog<<reflect1*Mmixing[h[0]][2]<<"\t"<<reflect1*Mmixing[h[0]][4]<<endl;
      clog<<reflect2*Mmixing[h[1]][2]<<"\t"<<reflect2*Mmixing[h[1]][4]<<endl;
      clog<<"and the one derived from alpha is:"<<endl;
      clog<<cos(alpha)<<"\t"<<sin(alpha)<<endl;
      clog<<-sin(alpha)<<"\t"<<cos(alpha)<<endl<<endl; 
      
      clog<<"The matrix of differences is:"<<endl;
      clog<<abs(reflect1*Mmixing[h[0]][2]-cos(alpha))<<"\t"<<abs(reflect1*Mmixing[h[0]][4]-sin(alpha))<<endl;
      clog<<abs(reflect2*Mmixing[h[1]][2]+sin(alpha))<<"\t"<<abs(reflect2*Mmixing[h[1]][4]-cos(alpha))<<endl<<endl;   
    }*/
  }
}

#ifdef HDECAYMODE
#ifdef HBMODE
bool RxSM::PrepareHiggsBoundsInput(){
  
  /////////////////////////////////////////////////////////
  //Passing couplings, masses and decay properties to HB //
  /////////////////////////////////////////////////////////

  compute_alpha();

  if(DM_candidate){
    CreateInputFileHdecayRsinglet(Cp.Mass[h[0]],Cp.Mass[h[1]],alpha,Phi[4],Cp.L[3],Cp.L[4],2);
  }
  else{
    CreateInputFileHdecayRsinglet(Cp.Mass[h[0]],Cp.Mass[h[1]],alpha,Phi[4],Cp.L[3],Cp.L[4],1);
  }
  
  HdecayCalcSing(HdecayHsing);
  if(DM_candidate && isNaNorInf(HdecayHsing["H1Width"])){
    std::cerr<<"Bad sHdecay output. This point will be rejected. The width is "<<HdecayHsing["H1Width"]<<endl;
    return false;
  }
  else if(isNaNorInf(HdecayHsing["H1Width"]*HdecayHsing["H2Width"])){
    std::cerr<<"Bad sHdecay output. This point will be rejected. The widthH1 is "<<HdecayHsing["H1Width"]<<" and width2 is "<<HdecayHsing["H2Width"]<<endl;
    return false;
  }

  //----------------------------------------------------------//

  for(int i=0;i!=nHzero;++i){
    //Call of Hdecay to compute all branching ratios
    
    double Mi2sq=Mmixing[h[i]][2]*Mmixing[h[i]][2];
    
    HiggsBoundsAnalysis.Mh[i]=Cp.Mass[h[i]];
    HiggsBoundsAnalysis.CP[i]=1;

    if(DM_candidate){
      //Branching ratios and total width
      HiggsBoundsAnalysis.BR_hjss[0] = HdecayHsing["BR(H1 -> s sbar)"];
      HiggsBoundsAnalysis.BR_hjcc[0] = HdecayHsing["BR(H1 -> c cbar)"];
      HiggsBoundsAnalysis.BR_hjbb[0] = HdecayHsing["BR(H1 -> b bbar)"];
      HiggsBoundsAnalysis.BR_hjtt[0] = HdecayHsing["BR(H1 -> t tbar)"];
      HiggsBoundsAnalysis.BR_hjmumu[0] = HdecayHsing["BR(H1 -> mu+ mu-)"];
      HiggsBoundsAnalysis.BR_hjtautau[0] = HdecayHsing["BR(H1 -> tau+ tau-)"];
      HiggsBoundsAnalysis.BR_hjWW[0] = HdecayHsing["BR(H1 -> W+ W-)"];
      HiggsBoundsAnalysis.BR_hjZZ[0] = HdecayHsing["BR(H1 -> Z Z)"];
      HiggsBoundsAnalysis.BR_hjZga[0]= HdecayHsing["BR(H1 -> Z gamma)"];
      HiggsBoundsAnalysis.BR_hjgaga[0] = HdecayHsing["BR(H1 -> gamma gamma)"];
      HiggsBoundsAnalysis.BR_hjgg[0] = HdecayHsing["BR(H1 -> g g)"];
      HiggsBoundsAnalysis.MhGammaTot[0] = HdecayHsing["H1Width"];  
    }
    else{
      if(h[i]==0){
	//Branching ratios and total width
	HiggsBoundsAnalysis.BR_hjss[0] = HdecayHsing["BR(H1 -> s sbar)"];
	HiggsBoundsAnalysis.BR_hjcc[0] = HdecayHsing["BR(H1 -> c cbar)"];
	HiggsBoundsAnalysis.BR_hjbb[0] = HdecayHsing["BR(H1 -> b bbar)"];
	HiggsBoundsAnalysis.BR_hjtt[0] = HdecayHsing["BR(H1 -> t tbar)"];
	HiggsBoundsAnalysis.BR_hjmumu[0] = HdecayHsing["BR(H1 -> mu+ mu-)"];
	HiggsBoundsAnalysis.BR_hjtautau[0] = HdecayHsing["BR(H1 -> tau+ tau-)"];
	HiggsBoundsAnalysis.BR_hjWW[0] = HdecayHsing["BR(H1 -> W+ W-)"];
	HiggsBoundsAnalysis.BR_hjZZ[0] = HdecayHsing["BR(H1 -> Z Z)"];
	HiggsBoundsAnalysis.BR_hjZga[0]= HdecayHsing["BR(H1 -> Z gamma)"];
	HiggsBoundsAnalysis.BR_hjgaga[0] = HdecayHsing["BR(H1 -> gamma gamma)"];
	HiggsBoundsAnalysis.BR_hjgg[0] = HdecayHsing["BR(H1 -> g g)"];
	HiggsBoundsAnalysis.MhGammaTot[0] = HdecayHsing["H1Width"];
      }
      else if (h[i]==1){
	//Branching ratios and total width
	HiggsBoundsAnalysis.BR_hjss[1] = HdecayHsing["BR(H2 -> s sbar)"];
	HiggsBoundsAnalysis.BR_hjcc[1] = HdecayHsing["BR(H2 -> c cbar)"];
	HiggsBoundsAnalysis.BR_hjbb[1] = HdecayHsing["BR(H2 -> b bbar)"];
	HiggsBoundsAnalysis.BR_hjtt[1] = HdecayHsing["BR(H2 -> t tbar)"];
	HiggsBoundsAnalysis.BR_hjmumu[1] = HdecayHsing["BR(H2 -> mu+ mu-)"];
	HiggsBoundsAnalysis.BR_hjtautau[1] = HdecayHsing["BR(H2 -> tau+ tau-)"];
	HiggsBoundsAnalysis.BR_hjWW[1] = HdecayHsing["BR(H2 -> W+ W-)"];
	HiggsBoundsAnalysis.BR_hjZZ[1] = HdecayHsing["BR(H2 -> Z Z)"];
	HiggsBoundsAnalysis.BR_hjZga[1]= HdecayHsing["BR(H2 -> Z gamma)"];
	HiggsBoundsAnalysis.BR_hjgaga[1] = HdecayHsing["BR(H2 -> gamma gamma)"];
	HiggsBoundsAnalysis.BR_hjgg[1] = HdecayHsing["BR(H2 -> g g)"];
	HiggsBoundsAnalysis.MhGammaTot[1] = HdecayHsing["H2Width"];
      }
      else{
	cerr<<"ScannerS ERROR: in bool RxSM::PrepareHiggsBoundsInput() bad h[i] ordering of higgs"<<endl;
	exit(EXIT_FAILURE);
      }
    }
    //Neutral Higgs hadronic input
    //LEP
    HiggsBoundsAnalysis.CS_lep_hjZ_ratio[i]=Mi2sq;
    HiggsBoundsAnalysis.CS_lep_bbhj_ratio[i]=Mi2sq; 
    HiggsBoundsAnalysis.CS_lep_tautauhj_ratio[i]=Mi2sq; 
    //Tevatron
    HiggsBoundsAnalysis.CS_tev_hj_ratio[i]=Mi2sq;
    HiggsBoundsAnalysis.CS_tev_hjb_ratio[i]=Mi2sq;
    HiggsBoundsAnalysis.CS_tev_hjW_ratio[i]=Mi2sq;
    HiggsBoundsAnalysis.CS_tev_hjZ_ratio[i]=Mi2sq;
    HiggsBoundsAnalysis.CS_tev_vbf_ratio[i]=Mi2sq; 
    HiggsBoundsAnalysis.CS_tev_tthj_ratio[i]=Mi2sq;
    //LHC7
    HiggsBoundsAnalysis.CS_lhc7_hj_ratio[i]=Mi2sq;
    HiggsBoundsAnalysis.CS_lhc7_hjb_ratio[i]=Mi2sq;
    HiggsBoundsAnalysis.CS_lhc7_hjW_ratio[i]=Mi2sq;
    HiggsBoundsAnalysis.CS_lhc7_hjZ_ratio[i]=Mi2sq;
    HiggsBoundsAnalysis.CS_lhc7_vbf_ratio[i]=Mi2sq;
    HiggsBoundsAnalysis.CS_lhc7_tthj_ratio[i]=Mi2sq;
    //LHC8
    HiggsBoundsAnalysis.CS_lhc8_hj_ratio[i]=Mi2sq;
    HiggsBoundsAnalysis.CS_lhc8_hjb_ratio[i]=Mi2sq;
    HiggsBoundsAnalysis.CS_lhc8_hjW_ratio[i]=Mi2sq;
    HiggsBoundsAnalysis.CS_lhc8_hjZ_ratio[i]=Mi2sq;
    HiggsBoundsAnalysis.CS_lhc8_vbf_ratio[i]=Mi2sq;
    HiggsBoundsAnalysis.CS_lhc8_tthj_ratio[i]=Mi2sq;
  }

  //branching ratios for higgs to 2 higgs decay
  if(DM_candidate){
    HiggsBoundsAnalysis.BR_hjinvisible[0] = HdecayHsing["BR(H1 -> H2 H2)"];
  }
  else{
    HiggsBoundsAnalysis.BR_hjhihi[0][1] = HdecayHsing["BR(H1 -> H2 H2)"];
    HiggsBoundsAnalysis.BR_hjhihi[1][0] = HdecayHsing["BR(H2 -> H1 H1)"];
  }
  
  //----------------------------------------------------------/*/
  
  return true;
}
#endif
#endif

void RxSM::PrintHeader() const{
  //NEEDS TO BE ADAPTED
  std::clog<<endl<<"***********************************************"<<endl;
  std::clog<<"*********** "<< scanners_version<<" ************"<<endl;
  std::clog<<"***********************************************" <<endl;
  std::clog<<endl<<"---------- SUMMARY INFO FOR THIS RUN ---------- "<<endl;
  std::clog<<endl<<"Potential Type: RxSM"<<endl;
  std::clog<<endl<<"V(H,S) = "<<endl;
  std::clog<<endl<< "The expanded potential is:"<<endl;
  std::clog<<endl<<(*this)<<endl;
  std::clog<<endl<<"The ranges of parameters to be scanned are:"<<endl;
  std::clog<<endl<<"Ranges for vacuum expectation values:"<<endl;
  for(size_t i=0;i!=Phi.size();++i){
    if(phi_bounds[i])
      std::clog<<"phi"<<i<<" in ["<<Phi.vgetmin(i)<<":"<<Phi.vgetmax(i)<<"]"<< endl;
  }
  std::clog <<endl<<"Ranges for couplings:"<< endl;
  for(size_t i=0;i!=Cp.L.size();++i){
    if(L_bounds[i])
      std::clog<<"L"<<i<<" in ["<<Cp.L.vgetmin(i)<<":"<<Cp.L.vgetmax(i)<<"]"<< endl;
  }
  std::clog <<endl<<"Ranges for physical state masses:"<< endl;
  for(size_t i=0;i!=Cp.Mass.size();++i){
    if(mass_bounds[i])
    std::clog<<"m"<<i<<" in ["<<Cp.Mass.vgetmin(i)<<":"<<Cp.Mass.vgetmax(i)<<"]"<<endl;
  }
  std::clog<<endl;
  std::clog<<"The seed you are using for the random number generator is: "<<r.getseed()<< endl<< endl;

}

bool RxSM::CheckBounds() const{
//unlike Potential::CheckBounds(), this only checks dependent parameters for which the range was explicity declared in the TwoRxSM input file. if they were not declared in input file (and were loaded of TwoRxSM.potential by default) mass_bounds and L_bounds will be false and the parameters will not be checked
  for(int i=0;i!=5;++i)
    if(mass_bounds[i] && (Cp.Mass[i]<Cp.Mass.Vecmin[i] || Cp.Mass[i]>Cp.Mass.Vecmax[i]))
      return false;
  for(int i=0;i!=5;++i)
    if(phi_bounds[i] && (Phi[i]<Phi.Vecmin[i] || Phi[i]>Phi.Vecmax[i]))
      return false;
  for(int i=0;i!=5;++i)
    if(L_bounds[i] && (Cp.L[i]<Cp.L.Vecmin[i] || Cp.L[i]>Cp.L.Vecmax[i]))
      return false;
      
  return true;
}

#ifdef MICROMEGASMODE
void RxSM::assignMOvalues(){
  
  //xSM2 thing
  
  ////// Dark matter
  ////// Relic density
  ForceUG=1;  // to Force Unitary Gauge assign 1 
 
  assignVal("LL0",Cp.L[0]);
  assignVal("LL1",Cp.L[1]);
  assignVal("LL2",Cp.L[2]);
  assignVal("LL3",Cp.L[3]);
  assignVal("LL4",Cp.L[4]);
  assignVal("LL5",Cp.L[5]);
  assignVal("LL6",Cp.L[6]);
  assignVal("LL7",Cp.L[7]);
  assignVal("LL8",Cp.L[8]);
  assignVal("LL9",Cp.L[9]);
  assignVal("LL10",Cp.L[10]);
  assignVal("LL11",Cp.L[11]);
  // mixing matrix
  assignVal("Mmix02",Mmixing[0][2]);
  assignVal("Mmix12",Mmixing[1][2]);
  assignVal("Mmix04",Mmixing[0][4]);
  assignVal("Mmix14",Mmixing[1][4]);
  // VEV
  assignVal("vs",Phi[4]);
  
  assignVal("MH0",Cp.Mass[0]);		
  assignVal("MH1",Cp.Mass[1]);		
  assignVal("MAs",Cp.Mass[2]);  
  char cdmName[10];
  int err=sortOddParticles(cdmName);
  if(err) { cerr<<"Can't calculate "<<cdmName<< endl; exit(-1);}
}
#endif

