#ifndef SCANNERSEXP_H
#define SCANNERSEXP_H
#include <cmath>
namespace exp_c{
  // general
  constexpr double pi = 3.141592654;
  // electroweak constants
  constexpr double alpha_0 = 7.2973525698e-3;
  constexpr double alpha_mz = 1/128.862;
  constexpr double e = std::sqrt(4*pi*alpha_0);
  constexpr double m_W = 80.385;
  constexpr double m_Z = 91.1876;
  constexpr double stw = std::sqrt(1-m_W*m_W/m_Z/m_Z);
  constexpr double ctw = std::sqrt(m_W*m_W/m_Z/m_Z);
  constexpr double Gf = 1.1663787e-5;
  constexpr double v = std::sqrt(1/std::sqrt(2)/Gf);
  //particle masses
  constexpr double m_e = 0.510998928e-3;
  constexpr double m_t = 172.5;
  constexpr double m_b = 4.92;
  constexpr double m_tau = 1.77682;

  //h125 signal strengths (1606.02266)
  constexpr double mu_VoF = 1.09;
  constexpr double mu_VoF_p = 0.36;
  constexpr double mu_VoF_m = 0.28;

  constexpr double mu_gamgam = 1.10;
  constexpr double mu_gamgam_p = 0.23;
  constexpr double mu_gamgam_m = 0.21;

  constexpr double mu_ZZ = 1.27;
  constexpr double mu_ZZ_p = 0.28;
  constexpr double mu_ZZ_m = 0.24;

  constexpr double mu_WW = 1.06;
  constexpr double mu_WW_p = 0.21;
  constexpr double mu_WW_m = 0.18;

  constexpr double mu_tautau = 1.05;
  constexpr double mu_tautau_p = 0.33;
  constexpr double mu_tautau_m = 0.27;

  constexpr double mu_bb = 0.64;
  constexpr double mu_bb_p = 0.37;
  constexpr double mu_bb_m = 0.28;

  //h125 mass
  constexpr double m_h = 125.09;

  //h125 BRs created with hdecay v 6.51, OMIT ELW=1, mh = 125.09 GeV
  constexpr double BRsm_gamgam = 0.2318E-02;
  constexpr double BRsm_ZZ = 0.2591E-01;
  constexpr double BRsm_WW = 0.2075;
  constexpr double BRsm_tautau = 0.6360E-01;
  constexpr double BRsm_bb = 0.5914;
}





#endif
